#include <winsock2.h>
#include <iphlpapi.h>
#include <stdio.h>
#include <stdlib.h>

#include "MachineId.h"
#include "Logger.h"

// Link with Iphlpapi.lib
#pragma comment(lib, "IPHLPAPI.lib")

#define WORKING_BUFFER_SIZE 15000
#define MAX_TRIES 3

#define MALLOC(x) HeapAlloc(GetProcessHeap(), 0, (x))
#define FREE(x) HeapFree(GetProcessHeap(), 0, (x))

PIP_ADAPTER_ADDRESSES findAdapter(PIP_ADAPTER_ADDRESSES pAddresses, DWORD type);
int generateMachineId(unsigned char PhysicalAddress[], int PhysicalAddressLength);

MachineId::MachineId(void)
{
}

MachineId::~MachineId(void)
{
}

// See: http://msdn.microsoft.com/en-us/library/windows/desktop/aa365915%28v=vs.85%29.aspx
int MachineId::getMachineId()
{
    DWORD dwSize = 0;
    DWORD dwRetVal = 0;

    unsigned int i = 0;

    // Set the flags to pass to GetAdaptersAddresses
    ULONG flags = GAA_FLAG_INCLUDE_PREFIX;

    // Unspecified address family (both). Other values: AF_INET, AF_INET6
    ULONG family = AF_UNSPEC;

    LPVOID lpMsgBuf = NULL;

    PIP_ADAPTER_ADDRESSES pAddresses = NULL;
    ULONG outBufLen = 0;
    ULONG Iterations = 0;

    PIP_ADAPTER_ADDRESSES pCurrAddresses = NULL;
    PIP_ADAPTER_UNICAST_ADDRESS pUnicast = NULL;
    PIP_ADAPTER_ANYCAST_ADDRESS pAnycast = NULL;
    PIP_ADAPTER_MULTICAST_ADDRESS pMulticast = NULL;
    IP_ADAPTER_DNS_SERVER_ADDRESS *pDnServer = NULL;
    IP_ADAPTER_PREFIX *pPrefix = NULL;
    // Allocate a 15 KB buffer to start with.
    outBufLen = WORKING_BUFFER_SIZE;

    do {
        pAddresses = (IP_ADAPTER_ADDRESSES *) MALLOC(outBufLen);
        if (pAddresses == NULL) {
            setErrorMessage("Memory allocation failed for IP_ADAPTER_ADDRESSES struct");
            Logger::logln("Memory allocation failed for IP_ADAPTER_ADDRESSES struct");
            return 0x0;
        }
        
        // The order in which adapters appear in the list returned by this function 
        // can be controlled from the Network Connections folder: select the Advanced Settings menu item from the Advanced menu
        dwRetVal = GetAdaptersAddresses(family, flags, NULL, pAddresses, &outBufLen);
        if (dwRetVal == ERROR_BUFFER_OVERFLOW) {
            FREE(pAddresses);
            pAddresses = NULL;
        } else {
            break;
        }
        Iterations++;
    } while ((dwRetVal == ERROR_BUFFER_OVERFLOW) && (Iterations < MAX_TRIES));

    int machineId = 0;
    if (dwRetVal == NO_ERROR) {
        
        // Ethernet card first
        pCurrAddresses = findAdapter(pAddresses, IF_TYPE_ETHERNET_CSMACD);
        
        // Wireless card next
        if (!pCurrAddresses) {
            pCurrAddresses = findAdapter(pAddresses, IF_TYPE_IEEE80211);
        }
        
        // Other network card
        if (!pCurrAddresses) {
            pCurrAddresses = findAdapter(pAddresses, -1);
        }
        
        if (pCurrAddresses) {
            machineId = generateMachineId(pCurrAddresses->PhysicalAddress, (int) pCurrAddresses->PhysicalAddressLength);
        } else {
            Logger::logln("Cannot find network adapter. Run ipconfig /all to list known adapters.");
        }

    } else {
        char messageBuffer [100];
        setErrorMessage(messageBuffer);
        Logger::logln(messageBuffer);
    }

    if (pAddresses) {
        FREE(pAddresses);
    }

    return machineId;
}

PIP_ADAPTER_ADDRESSES findAdapter(PIP_ADAPTER_ADDRESSES pAddresses, DWORD type)
{
    PWCHAR pName = NULL;
    PIP_ADAPTER_ADDRESSES pAddress = NULL;
    PIP_ADAPTER_ADDRESSES pCurrAddresses = pAddresses;

    while (pCurrAddresses) {
        if ((type == -1) || (pCurrAddresses->IfType == type)) {
            // Sort adapters by name (pr108045)
            if ((pCurrAddresses->PhysicalAddressLength != 0) && (pCurrAddresses->FriendlyName != NULL)) {
                if ((pName == NULL) || wcscmp(pCurrAddresses->FriendlyName, pName) < 0) {
                    pName = pCurrAddresses->FriendlyName;
                    pAddress = pCurrAddresses;
                }
            }
        }
        pCurrAddresses = pCurrAddresses->Next;
    }
    return pAddress;
}

int generateMachineId(unsigned char PhysicalAddress[], int PhysicalAddressLength)
{
    // Copied from Linux machine id calculation
    union U {
        unsigned char c[4];
        unsigned int  n;
    } u;

    unsigned int cSum = 0;
    for (int i = 0; i < PhysicalAddressLength; i++) {
        cSum += PhysicalAddress[i];
    }
    if (PhysicalAddressLength > 1)
        u.c[0] = PhysicalAddress[1];
    if (PhysicalAddressLength > 3)
        u.c[1] = PhysicalAddress[3];
    if (PhysicalAddressLength > 5)
        u.c[2] = PhysicalAddress[5]; 
    u.c[3] = (unsigned char)cSum;
    return u.n; 
}
